//-----------------------------------------------------------------------------
// File:	RacorX.cpp
//
// Desc:	Per-pixel lighting
//
//			This example uses the Common File routines.
//
//			Last modification: July 12th, 2002
//
// Copyright (c) 1995-2000 Microsoft Corporation. All rights reserved.
// Copyright (c) 2002 wolf@direct3d.net All rights reserved.
// Copyright (c) 1997 - 2002 NVIDIA Corporation. All rights reserverd.
//				I use the source of the following two functions: 
//					CreateBasisVectors()
//					FindAndFixDegenerateVertexBasis()
//				They are taken from the NVIDIA EffectsBrowser/Cg Browser
//			    Source	(Dot3_Util.cpp)
//-----------------------------------------------------------------------------

#define STRICT
#include <Windows.h>
#include <tchar.h>
#include <stdio.h>
#include <d3dx8.h>
#include "D3DApp.h"
#include "D3DFont.h"
#include "DXUtil.h"
#include <assert.h> 


#include "trackball.h"

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------
// temporary vertex structure to calculate the 
// tangent vector. This is used by CreateBasisVector()
struct TempVertex
{
    D3DXVECTOR3 Position;
    D3DXVECTOR3 Normal;
	D3DXVECTOR2 Texture;
	D3DXVECTOR3 S; 
	D3DXVECTOR3 SxT; 
	D3DXVECTOR3 T; 
};

// vertex buffer layout that is used by DIP()
struct ShaderVertex
{
    D3DXVECTOR3 Position;
 //   D3DXVECTOR3 Normal;
	D3DXVECTOR2 Texture;
	D3DXVECTOR3 S;   // tangent
	D3DXVECTOR3 SxT; // binormal
};


//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
public:
    CMyD3DApplication();
	~CMyD3DApplication();

    CD3DFont* m_pFont;
    CD3DFont* m_pFontSmall;

    D3DXMATRIX  m_matWorld;
    D3DXMATRIX  m_matView;
    D3DXMATRIX  m_matProj;
    D3DXMATRIX  m_matPosition;

    D3DXVECTOR4 m_LightDir;
    BOOL m_bLMouseDown;
    BOOL m_bRMouseDown;
    FLOAT m_xPos, m_yPos;
    FLOAT m_oPosX, m_oPosY;

    BYTE  m_bKey[256];
    FLOAT m_fTheta;
    FLOAT m_fTheta2;

    trackball_t m_pTrackball;	// trackball to turn the model
    bool m_bZoomDrag;			// zooming
    float m_fZoom;

    LPDIRECT3DVERTEXBUFFER8 m_pVertices;
    LPDIRECT3DINDEXBUFFER8	m_pIndexBuffer;
    INT m_iNumTriangles, m_iNumVertices;

    LPDIRECT3DTEXTURE8		m_pNormalMap;
    LPDIRECT3DTEXTURE8		m_pColorTexture;
    LPDIRECT3DTEXTURE8		m_pIllumMap;
	LPDIRECT3DTEXTURE8		m_pLightMap16;
	LPDIRECT3DTEXTURE8      m_pPointLightTexture;
	LPDIRECT3DCUBETEXTURE8	m_pCubeNormalizationMap;			// cube normalization map

    // shader handles
	DWORD m_dwPixelShader;
	DWORD m_dwVertexShader;

    TCHAR m_cColorMap[64];
    TCHAR m_cHeightMap[64];
	TCHAR m_cIllumMap[64];
	TCHAR m_cAttenuationmap[64];
	TCHAR m_cCubeNormalizationMap[64];

	TCHAR m_strStats[1024];	// help string

   
    virtual HRESULT OneTimeSceneInit();
    virtual HRESULT InitDeviceObjects();
    virtual HRESULT RestoreDeviceObjects();
    virtual HRESULT InvalidateDeviceObjects();
    virtual HRESULT DeleteDeviceObjects();
    virtual HRESULT FrameMove();
    virtual HRESULT Render();
    virtual LRESULT MsgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
    virtual HRESULT ConfirmDevice(D3DCAPS8*,DWORD,D3DFORMAT);
    virtual HRESULT FinalCleanup();

private:
    HRESULT LoadXFile(TCHAR*);
	HRESULT CreateVSFromCompiledFile (IDirect3DDevice8* m_pd3dDevice, 
													 DWORD* dwDeclaration,
													 TCHAR* strVSPath, 
													 DWORD* m_dwVS);
	HRESULT CreatePSFromCompiledFile (IDirect3DDevice8* m_pd3dDevice, 
													 TCHAR* strPSPath, 
													 DWORD* dwPS);
	
	// the following two functions calculate the tangent vectors
	HRESULT CreateBasisVectors(LPDIRECT3DVERTEXBUFFER8 pVertexBuffer, 
								LPDIRECT3DINDEXBUFFER8 pIndexBuffer);
	HRESULT FindAndFixDegenerateVertexBasis(TempVertex	* pVertices,
											DWORD		  dwNumVertices,
											void		* pIndices,
											DWORD		  dwNumIndices,
											bool bUnifyNormals,
											bool is32BitIndices);
};

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    CMyD3DApplication d3dApp;

    if( FAILED( d3dApp.Create( hInst ) ) )
        return 0;

    return d3dApp.Run();
}


//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
    if( (dwBehavior & D3DCREATE_HARDWARE_VERTEXPROCESSING ) ||
        (dwBehavior & D3DCREATE_MIXED_VERTEXPROCESSING ) )
    {
        if( pCaps->VertexShaderVersion < D3DVS_VERSION(1,1) )
            return E_FAIL;
    }
    if( pCaps->PixelShaderVersion < D3DPS_VERSION(1,4) )
         return E_FAIL;

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
    D3DXMATRIX matT;
    m_fElapsedTime			= 0.0f;
    m_dwCreationWidth		= 800;   // Width used to create window
    m_dwCreationHeight		= 600;
    m_bUseDepthBuffer		= TRUE;
    m_bShowCursorWhenFullscreen = TRUE;
    m_strWindowTitle		= _T("Per-Pixel Diffuse & Specular Lighting");
    m_pFont					= new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
    m_pFontSmall			= new CD3DFont( _T("Arial"),  9, D3DFONT_BOLD );
	memset (m_bKey, 0, sizeof (m_bKey));
    m_bLMouseDown			= FALSE;
    m_bRMouseDown			= FALSE;
    m_bZoomDrag				= false;
    m_fZoom					= -50.0f;			// initial zoom
    m_fTheta				= 0;
    m_fTheta2				= 0;
    m_LightDir				= D3DXVECTOR4( 0.0f, 0.0f, 1.0f, 1.0f );
	m_pVertices				= NULL;
	m_pIndexBuffer			= NULL;
	m_iNumTriangles			= 0;
	m_iNumVertices			= 0;
	m_pNormalMap			= NULL;
	m_pPointLightTexture	= NULL;
	m_pColorTexture			= NULL;
	m_pIllumMap				= NULL;
	m_pLightMap16			= NULL;
    D3DXMatrixIdentity(&m_matWorld);
    D3DXMatrixIdentity(&m_matPosition);
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application denstructor
//-----------------------------------------------------------------------------
CMyD3DApplication::~CMyD3DApplication()
{
	SAFE_DELETE(m_pFont);
	SAFE_DELETE(m_pFontSmall);
}


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: track mouse movement
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    // Trap the context menu
    if( WM_CONTEXTMENU==uMsg )
        return 0;

    switch(uMsg)
    {
    case WM_LBUTTONDOWN:
		SetCapture(hwnd);
        return 0;
    case WM_LBUTTONUP:
		ReleaseCapture();
        return 0;

	// track mouse movement to rotate sphere
    case WM_MOUSEMOVE:
	{
	    RECT r;
		FLOAT dx,dy;
		D3DXMATRIX mat1,mat2,mat3;

		GetWindowRect(m_hWnd,&r);
	    dx = ((FLOAT)m_xPos)/((FLOAT)r.right-r.left) * 2 - 1.0f;
	    dy = ((FLOAT)m_yPos)/((FLOAT)r.right-r.left) * 2 - 0.7f;
    
	    // left button rotates the object
	    if(m_bLMouseDown)
		{   
	      m_fTheta  = -dy+m_oPosY;
	      m_fTheta2 = -dx+m_oPosX;
    
	      D3DXMatrixRotationX(&mat1,m_fTheta * 0.1f);
	      D3DXMatrixRotationY(&mat2,m_fTheta2 * 0.1f);
	      D3DXMatrixMultiply(&mat3,&mat1,&mat2);
	      D3DXMatrixMultiply(&m_matWorld,&m_matWorld,&mat3);
		}
	}
    break;

    case WM_KEYUP:
        // mark a key up
        m_bKey[wParam] = FALSE;
        break;

    case WM_KEYDOWN:
        // mark a key down
         m_bKey[wParam] = TRUE;
       break;
   }

    return CD3DApplication::MsgProc( hwnd, uMsg, wParam, lParam );
}



//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Animates the scene
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{

	m_pTrackball.TrackMouse(m_hWnd, m_bZoomDrag, &m_fZoom); // track mouse movements and move object
    m_pTrackball.SetMatrix(m_matWorld); // get the world matrix for the trackball

    // setup view matrix
    D3DXVECTOR3 vUpVec(0,1,0);
    D3DXVECTOR3 vEyePt(0,0,m_fZoom);
    D3DXVECTOR3 vLookatPt(0,0,0);
    D3DXMatrixLookAtLH(&m_matView, &vEyePt, &vLookatPt, &vUpVec);

    if(m_bKey['W']) m_fZoom += 10.0f * m_fElapsedTime; // Move Forward
    if(m_bKey['S']) m_fZoom -= 10.0f * m_fElapsedTime; // Move Backward

	// move light
	if(m_bKey[VK_UP])	m_LightDir.y -= 1.5f  * m_fElapsedTime;
	if(m_bKey[VK_DOWN]) m_LightDir.y += 1.5f  * m_fElapsedTime;
	if(m_bKey[VK_LEFT])	m_LightDir.x += 1.5f  * m_fElapsedTime;
	if(m_bKey[VK_RIGHT]) m_LightDir.x -= 1.5f  * m_fElapsedTime;
	if(m_bKey[VK_HOME])	m_LightDir.z += 1.0f  * m_fElapsedTime;
	if(m_bKey[VK_END]) m_LightDir.z -= 1.0f  * m_fElapsedTime;
	if (m_LightDir.z >= 2.5f) m_LightDir.z = 2.5f; 
	if (m_LightDir.z <= 0.7f) m_LightDir.z = 0.7f; 

	// set light direction
    m_pd3dDevice->SetVertexShaderConstant(12, &m_LightDir,  1);	// light direction

	// world * view * proj matrix
    D3DXMATRIX  matTemp;
    D3DXMatrixTranspose(&matTemp,&(m_matWorld * m_matView * m_matProj));
    m_pd3dDevice->SetVertexShaderConstant(8, &matTemp, 4);	

	// world matrix
    D3DXMatrixTranspose(&matTemp,&m_matWorld);
    m_pd3dDevice->SetVertexShaderConstant(0, &matTemp, 4);	

	// eye vector
    m_pd3dDevice->SetVertexShaderConstant(24, &vEyePt, 1);		

   return S_OK;
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
    m_pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 
						D3DCOLOR_ARGB( 0x00, 0x00, 0x00, 0x00), 1.0, 0);
   
    // Begin the scene
    if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
    {
		m_pd3dDevice->SetTexture(0,m_pNormalMap); 
		m_pd3dDevice->SetTexture(2,m_pColorTexture);
		m_pd3dDevice->SetTexture(3,m_pIllumMap);
		m_pd3dDevice->SetTexture(4,m_pPointLightTexture);
		m_pd3dDevice->SetTexture(5,m_pCubeNormalizationMap);
			
	    m_pd3dDevice->SetPixelShader(m_dwPixelShader);
		m_pd3dDevice->SetVertexShader(m_dwVertexShader);

	    m_pd3dDevice->SetStreamSource( 0, m_pVertices, sizeof(ShaderVertex) );
		m_pd3dDevice->SetIndices(m_pIndexBuffer,0);
		m_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,m_iNumVertices,0,m_iNumTriangles);

		// Output statistics
		m_pFont->DrawText( 2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats );
		m_pFont->DrawText( 2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats );
		sprintf(m_strStats, 
				"W -     Zoom In \
				\nS -     Zoom Out \
				\nHome/End -     Moves Light -z/+z\
				\nUp/Down -     Moves Light -y/+y\
				\nLeft/Right -     Moves Light -x/+x");
		m_pFontSmall->DrawText( 2, 40, D3DCOLOR_ARGB(255,255,255,255), m_strStats);

        // End the scene.
        m_pd3dDevice->EndScene();
	}
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()

{
    // Set cursor to indicate that user can move the object with the mouse
#ifdef _WIN64
    SetClassLongPtr( m_hWnd, GCLP_HCURSOR, (LONG_PTR)LoadCursor( NULL, IDC_SIZEALL ) );
#else
    SetClassLong( m_hWnd, GCL_HCURSOR, (LONG)LoadCursor( NULL, IDC_SIZEALL ) );
#endif

    strcpy(m_cHeightMap, _T("media/normal.tga"));
    strcpy(m_cColorMap,_T("media/earth.tga"));
    strcpy(m_cIllumMap,_T("media/diffspec.dds"));
	strcpy(m_cAttenuationmap,_T("media/PointLight.dds"));
	strcpy(m_cCubeNormalizationMap,_T("media/normalize.dds"));

   return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
    m_pFont->InitDeviceObjects( m_pd3dDevice );
    m_pFontSmall->InitDeviceObjects( m_pd3dDevice );

    TCHAR strFile[160];
    DXUtil_FindMediaFile(strFile, _T("sphere.x"));
    LoadXFile(strFile);

	// load
	//  color map
	//  normal map
	//  illumination map
	//  point light map
	//  cube normalization map
	if(FAILED(D3DXCreateTextureFromFile(m_pd3dDevice, m_cColorMap,&m_pColorTexture)))
        return E_FAIL;
	if(FAILED(D3DXCreateTextureFromFile(m_pd3dDevice, m_cHeightMap,&m_pNormalMap)))
        return E_FAIL;
	if(FAILED(D3DXCreateTextureFromFile(m_pd3dDevice, m_cIllumMap,&m_pIllumMap)))
        return E_FAIL;
	if(FAILED(D3DXCreateTextureFromFile(m_pd3dDevice, m_cAttenuationmap, &m_pPointLightTexture)))
		return E_FAIL;
	if(FAILED(D3DXCreateCubeTextureFromFile(m_pd3dDevice, m_cCubeNormalizationMap, &m_pCubeNormalizationMap)))
		return E_FAIL;

	// vertex shader declaration
    DWORD dwDecl[] = 
    {
        D3DVSD_STREAM(0),
        D3DVSD_REG(0,   D3DVSDT_FLOAT3), // position
        D3DVSD_REG(7,   D3DVSDT_FLOAT2), // texture
        D3DVSD_REG(8,   D3DVSDT_FLOAT3), // tangent
        D3DVSD_REG(9,   D3DVSDT_FLOAT3), // binormal
        D3DVSD_END()
    };

	CreateVSFromCompiledFile (m_pd3dDevice, dwDecl, "vertexshader.vso", &m_dwVertexShader);
	CreatePSFromCompiledFile (m_pd3dDevice, "pixelshader.pso", &m_dwPixelShader);

   return S_OK;
}


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
    m_pFont->RestoreDeviceObjects();
    m_pFontSmall->RestoreDeviceObjects();

    // color or point light texture
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ADDRESSU,  D3DTADDRESS_CLAMP );
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ADDRESSV,  D3DTADDRESS_CLAMP );

    // normal map or point light texture
	m_pd3dDevice->SetTextureStageState( 1, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	m_pd3dDevice->SetTextureStageState( 1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	m_pd3dDevice->SetTextureStageState( 1, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);
	m_pd3dDevice->SetTextureStageState( 1, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
	m_pd3dDevice->SetTextureStageState( 1, D3DTSS_ADDRESSV, D3DTADDRESS_CLAMP);

	// cube normalization map
	m_pd3dDevice->SetTextureStageState( 2, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
	m_pd3dDevice->SetTextureStageState( 2, D3DTSS_ADDRESSV, D3DTADDRESS_CLAMP);
	m_pd3dDevice->SetTextureStageState( 2, D3DTSS_ADDRESSW, D3DTADDRESS_CLAMP);

    // illumination map
    m_pd3dDevice->SetTextureStageState( 3, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
	m_pd3dDevice->SetTextureStageState( 3, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	m_pd3dDevice->SetTextureStageState( 3, D3DTSS_ADDRESSU,  D3DTADDRESS_CLAMP );
	m_pd3dDevice->SetTextureStageState( 3, D3DTSS_ADDRESSV,  D3DTADDRESS_CLAMP );

    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE );
    m_pd3dDevice->SetRenderState( D3DRS_DITHERENABLE, TRUE );
    m_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );
    m_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
    
    D3DXMATRIX matProjection;
    D3DXMatrixPerspectiveFovLH(&m_matProj, 0.05f, 
								(FLOAT)m_d3dsdBackBuffer.Width /(FLOAT)m_d3dsdBackBuffer.Height, 
								1.0f, 10000.0f);

	// init trackball 
	m_pTrackball.Init(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height);

	// range in x
	D3DXVECTOR4 Range(0.6f, 0.5f, 0.0f, 0.0f);
	m_pd3dDevice->SetVertexShaderConstant(33, &Range, 1);

    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
    m_pFont->InvalidateDeviceObjects();
    m_pFontSmall->InvalidateDeviceObjects();
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
    m_pFont->DeleteDeviceObjects();
    m_pFontSmall->DeleteDeviceObjects();

    SAFE_RELEASE(m_pIndexBuffer);
    SAFE_RELEASE(m_pVertices);

    SAFE_RELEASE(m_pNormalMap);
    SAFE_RELEASE(m_pColorTexture);
    SAFE_RELEASE(m_pIllumMap);
	SAFE_RELEASE(m_pPointLightTexture);
	SAFE_RELEASE(m_pCubeNormalizationMap);

	if(m_dwVertexShader)
            m_pd3dDevice->DeleteVertexShader(m_dwVertexShader);
	if(m_dwPixelShader)
            m_pd3dDevice->DeletePixelShader(m_dwPixelShader);

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: LoadXFile()
// Desc: loads the x-file
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::LoadXFile(TCHAR* name)
{
    HRESULT		hr;
    LPD3DXMESH	pMeshSysMem = NULL,pMeshSysMem2 = NULL,pMeshSysMem3 = NULL;
	LPDIRECT3DVERTEXBUFFER8 pTempVertexBuffer = NULL;
	LPDIRECT3DINDEXBUFFER8 pTempIndexBuffer = NULL;
  
    if (FAILED (D3DXLoadMeshFromX(name, D3DXMESH_SYSTEMMEM, 
                          m_pd3dDevice, NULL, NULL, NULL, &pMeshSysMem)))
		return E_FAIL;

	// struct TempVertex
	// {
	//		D3DXVECTOR3 Position; // D3DFVF_XYZ
    //		D3DXVECTOR3 Normal;	  // D3DFVF_NORMAL
	//	
	//	D3DFVF_TEX4 four texture coordinates
	//		D3DXVECTOR2 Texture;  // tex coord. set 0
	//		D3DXVECTOR3 S;		  // tex coord. set 1 D3DFVF_TEXCOORDSIZE3(1)
	//		D3DXVECTOR3 SxT;	  // tex coord. set 2 D3DFVF_TEXCOORDSIZE3(2)
	//		D3DXVECTOR3 T;		  // tex coord. set 3 D3DFVF_TEXCOORDSIZE3(3)
	//	};
    hr = pMeshSysMem->CloneMeshFVF(D3DXMESH_MANAGED,  D3DFVF_XYZ | D3DFVF_NORMAL
                                    | D3DFVF_TEX4 | D3DFVF_TEXCOORDSIZE3(1)
									| D3DFVF_TEXCOORDSIZE3(2)
									| D3DFVF_TEXCOORDSIZE3(3), 
                                    m_pd3dDevice, &pMeshSysMem2);

    // compute the normals
/*    hr = D3DXComputeNormals(pMeshSysMem,NULL);
    if(FAILED(hr))
        return E_FAIL;

	// compute U (tangent)
	hr = D3DXComputeTangent(pMeshSysMem,0,pMeshSysMem2,1,D3DX_COMP_TANGENT_NONE,TRUE,NULL);
    if(FAILED(hr))
        return E_FAIL;
*/ 
    hr = pMeshSysMem2->GetIndexBuffer(&pTempIndexBuffer);
    if(FAILED(hr))
        return E_FAIL;
    hr = pMeshSysMem2->GetVertexBuffer(&pTempVertexBuffer);
    if(FAILED(hr))
        return E_FAIL;

	if (FAILED(CreateBasisVectors(pTempVertexBuffer, pTempIndexBuffer)))
		return E_FAIL;

    m_iNumTriangles = pMeshSysMem2->GetNumFaces();
    m_iNumVertices = pMeshSysMem2->GetNumVertices();

	// Two vectors are send to the vertex shader to build the
	// texture coordinate system. The third vector is build
	// in the vertex shader. This saves bandwidth.	
	// struct ShaderVertex
	// {
	//		D3DXVECTOR3 Position; // D3DFVF_XYZ
	//	
	//	D3DFVF_TEX3 three texture coordinates
	//		D3DXVECTOR2 Texture;  // tex coord. set 0
	//		D3DXVECTOR3 S		  // tex coord. set 1 D3DFVF_TEXCOORDSIZE3(1)
	//		D3DXVECTOR3 SxT		  // tex coord. set 1 D3DFVF_TEXCOORDSIZE3(1)
	//	};
	// This example doesn't use a normal, because we use a normal map
	// for lighting. 
	// Note: for N-Patches we need the Normal in input register
	// v3 in the vertex shader
    hr = pMeshSysMem2->CloneMeshFVF(D3DXMESH_MANAGED,  D3DFVF_XYZ 
                                    | D3DFVF_TEX3 | D3DFVF_TEXCOORDSIZE3(1)
									| D3DFVF_TEXCOORDSIZE3(2), 
                                    m_pd3dDevice, &pMeshSysMem3);

    hr = pMeshSysMem3->GetIndexBuffer(&m_pIndexBuffer);
    if(FAILED(hr))
        return E_FAIL;
    hr = pMeshSysMem3->GetVertexBuffer(&m_pVertices);
    if(FAILED(hr))
        return E_FAIL;
    
    // cleanup
    SAFE_RELEASE(pMeshSysMem)
    SAFE_RELEASE(pMeshSysMem2)
    SAFE_RELEASE(pMeshSysMem3)
	SAFE_RELEASE(pTempVertexBuffer)
	SAFE_RELEASE(pTempIndexBuffer)

  return S_OK;
}


//------------------------------------------------------------------------------
// Name:	CreateVSFromBinFile
// Desc:	loads a binary *.vso file 
//			and creates a vertex shader
//------------------------------------------------------------------------------
HRESULT CMyD3DApplication::CreateVSFromCompiledFile (IDirect3DDevice8* m_pd3dDevice, 
													 DWORD* dwDeclaration,
													 TCHAR* strVSPath, 
													 DWORD* m_dwVS)
{
	char szBuffer[128];		// debug output
	DWORD*	dwpVS;			// pointer to address space of the calling process
	HANDLE hFile, hMap;		// handle file and handle mapped file
    TCHAR tchTempVSPath[512];	// temporary file path
	HRESULT hr;				// error 

    if( FAILED( hr = DXUtil_FindMediaFile( tchTempVSPath, strVSPath ) ) )
        return D3DAPPERR_MEDIANOTFOUND;
	
	hFile = CreateFile(tchTempVSPath, GENERIC_READ,0,0,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,0);

	if(hFile != INVALID_HANDLE_VALUE) 
	{
		if(GetFileSize(hFile,0) > 0) 
			hMap = CreateFileMapping(hFile,0,PAGE_READONLY,0,0,0);
		else
		{
			CloseHandle(hFile);
			return E_FAIL;		
		}
	}	
	else
		return E_FAIL;	
	
	// maps a view of a file into the address space of the calling process
	dwpVS = (DWORD *)MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
		
	// Create the vertex shader
	hr = m_pd3dDevice->CreateVertexShader( dwDeclaration, dwpVS, m_dwVS, 0 );
	if ( FAILED(hr) )
	{
		OutputDebugString( "Failed to create Vertex Shader, errors:\n" );
		D3DXGetErrorStringA(hr,szBuffer,sizeof(szBuffer));
		OutputDebugString( szBuffer );
		OutputDebugString( "\n" );
	  return hr;
	}
	
	UnmapViewOfFile(dwpVS);
	CloseHandle(hMap);
	CloseHandle(hFile);
	
  return S_OK;
}

//------------------------------------------------------------------------------
// Name:	CreatePSFromBinFile
// Desc:	loads a binary *.pso file 
//			and creates a pixel shader
//------------------------------------------------------------------------------
HRESULT CMyD3DApplication::CreatePSFromCompiledFile (IDirect3DDevice8* m_pd3dDevice, 
													 TCHAR* strPSPath, 
													 DWORD* dwPS)
{
	char szBuffer[128];		// debug output
	DWORD*	pdwPS;			// pointer to address space of the calling process
	HANDLE hFile, hMap;		// handle file and handle mapped file
    TCHAR tchTempVSPath[512];	// temporary file path
	HRESULT hr;				// error 

    if( FAILED( hr = DXUtil_FindMediaFile( tchTempVSPath, strPSPath ) ) )
        return D3DAPPERR_MEDIANOTFOUND;
	
	hFile = CreateFile(tchTempVSPath, GENERIC_READ,0,0,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,0);

	if(hFile != INVALID_HANDLE_VALUE) 
	{
		if(GetFileSize(hFile,0) > 0) 
			hMap = CreateFileMapping(hFile,0,PAGE_READONLY,0,0,0);
		else
		{
			CloseHandle(hFile);
			return E_FAIL;		
		}
	}	
	else
		return E_FAIL;	
	
	// maps a view of a file into the address space of the calling process
	pdwPS = (DWORD *)MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
		
	// Create the pixel shader
	hr = m_pd3dDevice->CreatePixelShader(pdwPS, dwPS);
	if ( FAILED(hr) )
	{
		OutputDebugString( "Failed to create Pixel Shader, errors:\n" );
		D3DXGetErrorStringA(hr,szBuffer,sizeof(szBuffer));
		OutputDebugString( szBuffer );
		OutputDebugString( "\n" );
	  return hr;
	}
	
	UnmapViewOfFile(pdwPS);
	CloseHandle(hMap);
	CloseHandle(hFile);
	
  return S_OK;
}


//----------------------------------------------------------------------------
// The source of the following two functions 
//  CreateBasisVectors()
//  FindAndFixDegenerateVertexBasis()
// is taken from the NVIDIA EffectsBrowser Source
// I found it in the file Dot3_Util.cpp
//
// They are used to replace the D3DXComputeTangent() function provided with
// the Direct3D Utility library in the DirectX 8.1 SDK.
// Unfortunenately this function has a bug, that leads to less precision
// under certain circumstances. The lack of precision showed up, when using 
// the cube normalization map.
// This bug is a known bug and should be removed in DX 9. 
//-----------------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////////
// Creates basis vectors, based on a vertex and index list.
// NOTE: Assumes an indexed triangle list, with vertices of type Dot3Vertex
// Does not check for degenerate vertex positions - ie vertices with same 
//  position but different texture coords or normals.  Another function 
//  can do this to average the basis vectors & "smooth" the tangent space
//  for those duplicated vertices

#define SMALL_FLOAT 1e-12

HRESULT CMyD3DApplication::CreateBasisVectors(LPDIRECT3DVERTEXBUFFER8 pVertexBuffer, LPDIRECT3DINDEXBUFFER8 pIndexBuffer)
{
	DWORD i;
	TempVertex* pVertices;
	void* pIndices;
	HRESULT hr;
	
	assert(pVertexBuffer);
	assert(pIndexBuffer);

	// Calculate number of vertices and indices
	D3DVERTEXBUFFER_DESC VBDesc;
	D3DINDEXBUFFER_DESC IBDesc;
	
	pVertexBuffer->GetDesc(&VBDesc);
	pIndexBuffer->GetDesc(&IBDesc);

	DWORD dwNumIndices;
	DWORD dwNumVertices;
	switch(IBDesc.Format)
	{
		case D3DFMT_INDEX16:
			dwNumIndices = IBDesc.Size / 2;
			break;
		case D3DFMT_INDEX32:
			dwNumIndices = IBDesc.Size / 4;
			break;
		default:
			assert(0);
			return E_FAIL;
	}

	dwNumVertices = VBDesc.Size / sizeof(TempVertex);

	// Get a pointer to the indices and the vertices
	hr = pVertexBuffer->Lock(0, 0, (BYTE**)&pVertices, 0);
	if (FAILED(hr))
		return hr;

	hr = pIndexBuffer->Lock(0, 0, (BYTE**)&pIndices, 0);
	if (FAILED(hr))
		return hr;

	// Clear the basis vectors
	for (i = 0; i < dwNumVertices; i++)
	{
		pVertices[i].S = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
		pVertices[i].T = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	}

	// Walk through the triangle list and calculate gradiants for each triangle.
	// Sum the results into the S and T components.
    for( i = 0; i < dwNumIndices; i += 3 )
    {       
		DWORD TriIndex[3];
		D3DXVECTOR3 du, dv;
		D3DXVECTOR3 edge01;
		D3DXVECTOR3 edge02;
		D3DXVECTOR3 cp;
		
		switch(IBDesc.Format)
		{
		case D3DFMT_INDEX16:
			TriIndex[0] = ((WORD *)pIndices)[i];
			TriIndex[1] = ((WORD *)pIndices)[i+1];
			TriIndex[2] = ((WORD *)pIndices)[i+2];
			break;
		case D3DFMT_INDEX32:
			TriIndex[0] = ((DWORD *)pIndices)[i];
			TriIndex[1] = ((DWORD *)pIndices)[i+1];
			TriIndex[2] = ((DWORD *)pIndices)[i+2];
			break;
		}

		assert((TriIndex[0] < dwNumVertices) && (TriIndex[1] < dwNumVertices) && (TriIndex[2] < dwNumVertices));

		TempVertex& v0 = pVertices[TriIndex[0]];
		TempVertex& v1 = pVertices[TriIndex[1]];
		TempVertex& v2 = pVertices[TriIndex[2]];

		// x, s, t
		edge01 = D3DXVECTOR3( v1.Position.x - v0.Position.x, v1.Texture.x - v0.Texture.x, v1.Texture.y - v0.Texture.y );
		edge02 = D3DXVECTOR3( v2.Position.x - v0.Position.x, v2.Texture.x - v0.Texture.x, v2.Texture.y - v0.Texture.y );

		D3DXVec3Cross(&cp, &edge01, &edge02);
		if ( fabs(cp.x) > SMALL_FLOAT )
		{
			v0.S.x += -cp.y / cp.x;
			v0.T.x += -cp.z / cp.x;

			v1.S.x += -cp.y / cp.x;
			v1.T.x += -cp.z / cp.x;
			
			v2.S.x += -cp.y / cp.x;
			v2.T.x += -cp.z / cp.x;
		}

		// y, s, t
		edge01 = D3DXVECTOR3( v1.Position.y - v0.Position.y, v1.Texture.x - v0.Texture.x, v1.Texture.y - v0.Texture.y );
		edge02 = D3DXVECTOR3( v2.Position.y - v0.Position.y, v2.Texture.x - v0.Texture.x, v2.Texture.y - v0.Texture.y );

		D3DXVec3Cross(&cp, &edge01, &edge02);
		if ( fabs(cp.x) > SMALL_FLOAT )
		{
			v0.S.y += -cp.y / cp.x;
			v0.T.y += -cp.z / cp.x;

			v1.S.y += -cp.y / cp.x;
			v1.T.y += -cp.z / cp.x;
			
			v2.S.y += -cp.y / cp.x;
			v2.T.y += -cp.z / cp.x;
		}

		// z, s, t
		edge01 = D3DXVECTOR3( v1.Position.z - v0.Position.z, v1.Texture.x - v0.Texture.x, v1.Texture.y - v0.Texture.y );
		edge02 = D3DXVECTOR3( v2.Position.z - v0.Position.z, v2.Texture.x - v0.Texture.x, v2.Texture.y - v0.Texture.y );

		D3DXVec3Cross(&cp, &edge01, &edge02);
		if ( fabs(cp.x) > SMALL_FLOAT )
		{
			v0.S.z += -cp.y / cp.x;
			v0.T.z += -cp.z / cp.x;

			v1.S.z += -cp.y / cp.x;
			v1.T.z += -cp.z / cp.x;
			
			v2.S.z += -cp.y / cp.x;
			v2.T.z += -cp.z / cp.x;
		}
    }


	////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////
	// Before normalizing the S, T vectors we have to 
	//  find degenerate vertices to average the basis
	//  for points which share position but not UV
	//  coordinate (ie lathed objects with a seam where
	//  vertices are duplicated).
	// If we were to search for these degenerate points after 
	//  normalizing the S & T, the averaging of these normalized
	//  (pre-averaged) vectors would not be correct over the
	//  seam, and a discontinuity might appear.

		
	FindAndFixDegenerateVertexBasis( pVertices, dwNumVertices, pIndices, dwNumIndices, 
		true, (IBDesc.Format == D3DFMT_INDEX32));


	////////////////////////////////////////////////
	////////////////////////////////////////////////

    // Calculate the SxT vector
  	for(i = 0; i < dwNumVertices; i++)
  	{		
  		// Normalize the S, T vectors
  		D3DXVec3Normalize(&pVertices[i].S, &pVertices[i].S);
  		D3DXVec3Normalize(&pVertices[i].T, &pVertices[i].T);
  
  		// Get the cross of the S and T vectors
  		D3DXVec3Cross(&pVertices[i].SxT, &pVertices[i].S, &pVertices[i].T);
  
  		// Need a normalized normal
  		D3DXVec3Normalize(&pVertices[i].Normal, &pVertices[i].Normal);

  		// Get the direction of the SxT vector
  		if (D3DXVec3Dot(&pVertices[i].SxT, &pVertices[i].Normal) < 0.0f)
  		{
  			pVertices[i].SxT = -pVertices[i].SxT;
  		}
  	}

	pVertexBuffer->Unlock();
	pIndexBuffer->Unlock();

	return S_OK;
}




////////////////////////////////////////////////////////////////////////////
// FindAndFixDegenerateVertexBasis(..)
//  Takes a vertex buffer with vertices of type Dot3Vertex.  Indices must be
//  for an INDEXED TRIANGLE LIST.
// This finds vertices within epsilon in position and averages their tangent
//  bases to make a smooth tangent space across the model.  This is useful for
//  lathed objects or authored models which duplicate vertices along material
//  boundaries.
// Tangent Basis must have already been computed for this to work! =)
//
// If bUnifyNormals is true, then the vertex normals are averaged too
//

#define WORD_MAX	65535

HRESULT CMyD3DApplication::FindAndFixDegenerateVertexBasis(	TempVertex	* pVertices,
											DWORD		  dwNumVertices,
											void		* pIndices,
											DWORD		  dwNumIndices,
											bool bUnifyNormals,
											bool is32BitIndices)

{

	assert( pVertices != NULL );
	assert( pIndices != NULL );


	float epsilon = 1.0e-5f;
	float x,y,z,dist;

	////////////////////////////////////////////////////////////////

	DWORD i,j;

	////////////////////////////////////////////////////////////////
	// Sloppy, but allocate a pointer and char for each vertex
	// As overlapping vertices are found, increment their duplicate_count
	//   and allocate an array of MAX_OVERLAP vertex indices to reference
	//   which vertices overlap.

	#define MAX_OVERLAP 50

	BYTE * duplicate_count = new BYTE[ dwNumVertices ];
		// duplicate_index is array of pointers to bins.  Each bin is
		// allocated when a match is found.
	WORD ** duplicate_index = (WORD**) calloc( dwNumVertices, sizeof(WORD*) );

	memset( duplicate_count, 0, dwNumVertices * sizeof( BYTE  ) );
	memset( duplicate_index, 0, dwNumVertices * sizeof( WORD* ) );


	// Need to search the mesh for vertices with the same spatial coordinate
	// These are vertices duplicated for lathed/wrapped objects to make a
	//   2nd set of texture coordinates at the point in order to avoid bad
	//   texture wrapping
	// In short:  For each vertex, find any other vertices that share it's 
	//   position.  "Average" the tangent space basis calculated above for
	//   these duplicate vertices.  This is not rigorous, but works well 
	//   to fix up authored models.  ** Models should not have T juntions! **

	// Check each vert with every other.  There's no reason to check
	//   j with i after doing i with j, so start j from i ( so we test
	//   1 with 2 but never 2 with 1 again).
	// This is a model pre-processing step and only done once.  For large
	//   models, compute this off-line if you have to and store the resultant
	//   data.
	// The whole thing could be made much more efficient (linked list, etc)

	for( i=0; i < dwNumVertices; i++ )
	{
		for(j=i+1; j < dwNumVertices; j++ )
		{
			x = pVertices[i].Position.x - pVertices[j].Position.x;
			y = pVertices[i].Position.y - pVertices[j].Position.y;
			z = pVertices[i].Position.z - pVertices[j].Position.z;

			dist = x*x + y*y + z*z;
			
			if( dist < epsilon )
			{
				// if i matches j and k, just record into i.  j will be 
				//  half full as it will only match k, but this is
				//  taken care of when i is processed.
				if( duplicate_count[i] == 0 )
				{
					// allocate bin
					duplicate_index[i] = new WORD[MAX_OVERLAP];
				}
				if( duplicate_count[i] < MAX_OVERLAP )
				{
					assert( j < WORD_MAX );
					duplicate_index[i][duplicate_count[i]] = (WORD) j;
					duplicate_count[i] ++;
				}
				else
				{
	//				FDebug("Ran out of bin storage!!\n");
					assert( false );
				}
			}
		}

	}

	// Now average the tangent spaces & write the new result to
	//  each duplicated vertex


	D3DXVECTOR3	S_temp, T_temp, SxT_temp, N_temp;


	for( i = 0; i < dwNumVertices; i++ )
	{
		// do < 10 check to not average the basis at poles of sphere or
		//  other ridiculous geometry with too many degenerate vertices

		if( duplicate_count[i] > 0 && duplicate_count[i] < 10 )
		{
			//	FDebug("Averaging vert prop at %d for %d vertices\n", i, duplicate_count[i]);

			// If there are more than 5 vertices sharing this point then
			//  the point is probably some kind of lathe axis node.  No need to
			//  process it here

			// Set accumulator to value for vertex in question

			S_temp		= pVertices[i].S;
			T_temp		= pVertices[i].T;
			SxT_temp	= pVertices[i].SxT;
			N_temp		= pVertices[i].Normal;

			// add in basis vectors for all other vertices which
			//  have the same positon (found above)

			for( j=0; j < duplicate_count[i]; j++ )
			{
				S_temp		= S_temp   + pVertices[duplicate_index[i][j]].S;
				T_temp		= T_temp   + pVertices[duplicate_index[i][j]].T;
				SxT_temp	= SxT_temp + pVertices[duplicate_index[i][j]].SxT;

				N_temp		= N_temp   + pVertices[duplicate_index[i][j]].Normal;
			}

			// Normalize the basis vectors
			// Note that SxT might not be perpendicular to S and T
			//  anymore.  Not absolutely necessary to re-do the 
			//  cross product.

			D3DXVec3Normalize( & S_temp, & S_temp );
			D3DXVec3Normalize( & T_temp, & T_temp );
			D3DXVec3Normalize( & SxT_temp, & SxT_temp );
			D3DXVec3Normalize( & N_temp, & N_temp );

			// Write the average basis to the first vertex for which
			//   the duplicates were found

			pVertices[i].S = S_temp;
			pVertices[i].T = T_temp;
			pVertices[i].SxT = SxT_temp;

			if( bUnifyNormals )
				pVertices[i].Normal = N_temp;

			// Now write to all later vertices with the same position
				
			for(j=0; j < duplicate_count[i]; j++ )
			{
				// Set the vertices in the same position to
				//  the average basis.

				pVertices[duplicate_index[i][j]].S = S_temp;
				pVertices[duplicate_index[i][j]].T = T_temp;
				pVertices[duplicate_index[i][j]].SxT = SxT_temp;

				if( bUnifyNormals )
					pVertices[duplicate_index[i][j]].Normal = N_temp;


				// Kill the duplicate index lists of all vertices of
				//  higher index which overlap this one.  This is so
				//  higher index vertices do not average a smaller 
				//  subset of bases.
				// Arrays are de-allocated later

				duplicate_count[ duplicate_index[i][j] ] = 0;

			}

		}

		if( duplicate_index[i] != NULL )
		{
			delete [] duplicate_index[i];
			duplicate_index[i] = NULL;
			duplicate_count[i] = 0;
		}
	}

	delete [] duplicate_count;
	free( duplicate_index );


	return( S_OK );
}

