; Constant registers 
; c0-c3 - world space transposed
; c8-c11 - world * view * proj
; c12 - Light Direction (In World Space)
; c24 - camera direction
; c33 - range value
;
; Input registers
; v0 - Position
; v7 - Texture 
; v8 - Tangent
; v9 - SxT
vs.1.1

; position in clip space
m4x4 oPos, v0, c8

m3x3 r3, v8, c0	; generate tangent U
m3x3 r5, v9, c0	; generate normal W

; Cross product
; generate binormal V
mul r0, -r3.zxyw, r5.yzxw;
mad r4, -r3.yzxw, r5.zxyw,-r0;

; position in world space
m4x4 r2, v0, c0			

; get light vector
add r10, r2, -c12

;transform the light vector with U, V, W
dp3 oT1.x, r3, r10
dp3 oT1.y, r4, r10
dp3 oT1.z, r5, r10

; get a vector toward the camera/eye V
add r6, -r10, c24

add r6.xyz, r6.xyz, -c12		; get half angle

; transform the half angle vector into texture space
dp3 oT2.x, r3 ,r6
dp3 oT2.y, r4 ,r6
dp3 oT2.z, r5 ,r6

mov oT0.xy, v7.xy ; color map

; attenuation stuff
; get light vector
add r10, r2, c12

; point light stuff
; Divide each component by the range value
mul r10, r10, c33.x

; multiply with 0.5 and add 0.5
mad r10, r10, c33.yyyy, c33.yyyy

; map the x and y components into the first texture
mov oT4.xy, r10.xy

; z-component 
mov oT5.xy, r10.z
